from projectreport.analyzer.parsers.javascript import PackageJSONParser
from projectreport.analyzer.parsers.python import PythonParser

PARSER_EXTENSIONS = {"py": PythonParser}

PARSER_DOC_FILES = {
    "__init__.py": PythonParser,
    "package.json": PackageJSONParser,
}



import os
import re
from pathlib import Path
from typing import (
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from projectreport.searcher.rotating_list import RotatingList
from projectreport.tools.expand_glob import all_possible_paths


def read_all_files_in_folders_print_lines_around_regex(
    folders: Sequence[str],
    str_pattern: str,
    num_lines: int = 2,
    recursive: bool = False,
    print_lines: bool = True,
    ignore_paths: Optional[Sequence[str]] = None,
) -> Dict[str, List[List[str]]]:
    """
    Searches for regex in each line of every file in multiple folders. When regex is
    matched, will return num_lines lines around the file

    :param folders: Folders to search in
    :param str_pattern: Pattern to match in a line
    :param num_lines: Number of lines around the matching line to return/print,
        defaults to 2
    :param recursive: Whether to search folders within passed folder, defaults to False
    :param print_lines: Whether to print the results, defaults to False
    :param ignore_paths: Relative paths to ignore. Globs are accepted, defaults to None
    :return: a dictionary where keys are file paths, and values are lists where each
        element is a list containing the lines for one match
    """
    found_lines = {}
    for folder in folders:
        found_lines.update(
            read_all_files_in_folder_print_lines_around_regex(
                folder,
                str_pattern,
                num_lines=num_lines,
                recursive=recursive,
                print_lines=print_lines,
                ignore_paths=ignore_paths,
            )
        )
    return found_lines


def read_all_files_in_folder_print_lines_around_regex(
    file_path: str,
    str_pattern: str,
    num_lines: int = 2,
    recursive: bool = False,
    print_lines: bool = True,
    ignore_paths: Optional[Sequence[str]] = None,
) -> Dict[str, List[List[str]]]:
    """
    Searches for regex in each line of every file in a folder. When regex is matched,
    will return num_lines lines around the file

    :param file_path: Path of file to search in
    :param str_pattern: Pattern to match in a line
    :param num_lines: Number of lines around the matching line to return/print,
        defaults to 2
    :param recursive: Whether to search folders within passed folder, defaults to False
    :param print_lines: Whether to print the results, defaults to False
    :param ignore_paths: Relative paths to ignore. Globs are accepted, defaults to None
    :return: a dictionary where keys are file paths, and values are lists where each
        element is a list containing the lines for one match
    """

    iterator: Union[
        List[Tuple[str, List[str], List[str]]],
        Iterator[Tuple[str, List[str], List[str]]],
    ]
    if recursive:
        iterator = os.walk(file_path)
    else:
        iterator = [next(os.walk(file_path))]

    if ignore_paths is None:
        ignore_paths = []

    all_absolute_ignore_paths: Set[Path] = set()
    found_lines = {}

    def should_ignore_path(path_str: str) -> bool:
        path = Path(path_str)
        for ignore_path in all_absolute_ignore_paths:
            if path == ignore_path or ignore_path in path.parents:
                return True
        return False

    for path, folders, files in iterator:
        expanded_ignore_paths = all_possible_paths(ignore_paths, path)
        all_absolute_ignore_paths.update(
            set([Path(path_str) for path_str in expanded_ignore_paths])
        )
        if should_ignore_path(path):
            # Skip ignored folder
            continue
        for file in files:
            full_path = os.path.join(path, file)
            if should_ignore_path(full_path):
                # Skip ignored file
                continue
            lines = read_file_get_lines_around_regex(
                full_path, str_pattern, num_lines=num_lines, print_lines=False
            )
            if lines:
                found_lines[full_path] = lines
                if print_lines:
                    print(f"\n\nFound {len(lines)} match in {full_path}")
                    for line_set in lines:
                        _print_tracked_lines(line_set)

    return found_lines


def read_file_get_lines_around_regex(
    file_path: str, str_pattern: str, num_lines: int = 2, print_lines: bool = False
) -> List[List[str]]:
    """
    Searches for regex in each line of a file. When regex is matched, will return
    num_lines lines around the file

    :param file_path: path of file to search in
    :param str_pattern: pattern to match in a line
    :param num_lines: number of lines around the matching line to return/print,
        defaults to 2
    :param print_lines: whether to print the results, defaults to False
    :return: a list where each element is a list containing the lines for one match
    """
    reader = _file_reader(file_path)
    lines = _get_lines_around_regex(
        str_pattern, reader, num_lines=num_lines, print_lines=print_lines
    )
    return lines


def _file_reader(filename):
    try:
        with open(filename, "r", encoding="utf8") as f:
            for line in f:
                yield line.strip()
    except UnicodeDecodeError:
        try:
            with open(filename, "r", encoding="latin1") as f:
                for line in f:
                    yield line.strip()
        except Exception as e:
            print(f"Could not read file {filename}: {e}")
            return


def _get_lines_around_regex(
    str_pattern: str,
    lines: Generator[str, None, None],
    num_lines: int = 2,
    print_lines: bool = False,
) -> List[List[str]]:
    pattern = re.compile(str_pattern)

    total_num_lines = (
        num_lines * 2 + 1
    )  # 1 line for match, then 2 * num_lines for before and after match

    # After finding a match, must delay the print for num_lines as only
    # then will we have the lines after
    # the match. Set up this list to track at which lines we should print
    print_at_lines = []

    # Keeps only the last total_num_lines entires
    tracked_lines = RotatingList([], total_num_lines)

    # Will hold each set of found lines to return at the end
    found_lines = []

    def record_lines():
        found_lines.append(list(tracked_lines))
        if print_lines:
            _print_tracked_lines(tracked_lines)

    # Find and track lines which have matches, printing num_lines after
    for i, line in enumerate(lines):
        line_num = i + 1
        tracked_lines.append(f"{line_num}: {line}")
        if _matches_regex(pattern, line):
            print_at_lines.append(line_num + num_lines)
        if line_num in print_at_lines:
            record_lines()

    # Print final section if requested line to print at was after the end of the file
    if any([print_line_num > line_num for print_line_num in print_at_lines]):
        record_lines()

    return found_lines


def _print_tracked_lines(lines: Union[RotatingList, List[str]]) -> None:
    print("\n" + "\n".join(lines))


def _matches_regex(pattern: re.Pattern, search_str: str) -> bool:
    match = re.search(pattern, search_str)
    return match is not None



from typing import Optional, Sequence

from projectreport import Project, Report
from tests.fixtures.project import get_python_project


def get_project_report(add_projects: Optional[Sequence[Project]] = None) -> Report:
    py_project = get_python_project()

    projects = [
        py_project,
    ]

    if add_projects:
        projects.extend(add_projects)

    report = Report(projects)
    return report



from typing import TYPE_CHECKING, Dict, Union

from projectreport.data import AnalysisData

if TYPE_CHECKING:
    from projectreport.analyzer.project import Project
    from projectreport.analyzer.folder import Folder

from cached_property import cached_property

from projectreport.report.base import BaseReport


class ProjectReport(BaseReport):
    def __init__(self, project: "Project", depth: int = 0):
        self.project = project
        self.depth = depth

    @cached_property
    def doc(self):
        from projectreport.report.latex import project_latex_document

        return project_latex_document(self.data)

    @cached_property
    def data(self) -> AnalysisData:
        data: AnalysisData = {}
        data.update(self.project.data)
        if self.depth == 0:
            return data

        subprojects_dict: AnalysisData = {}
        for folder in self.project.folders:
            folder_data: AnalysisData = {}
            gather_data(folder, folder_data, remaining_depth=self.depth - 1)
            subprojects_dict[folder.path] = folder_data
        data.update({"subprojects": subprojects_dict})

        return data


def gather_data(folder: "Folder", data: dict, remaining_depth: int = 0):
    data.update(folder.data)
    if remaining_depth <= 0:
        return

    subprojects_dict: AnalysisData = {}
    for folder in folder.folders:
        folder_data: AnalysisData = {}
        gather_data(folder, folder_data, remaining_depth=remaining_depth - 1)
        subprojects_dict[folder.path] = folder_data
    data.update({"subprojects": subprojects_dict})



import datetime
import os
from copy import deepcopy

import pytz

from tests.config import REPORTS_FOLDER
from tests.fixtures.project import get_js_project, git_project
from tests.fixtures.report import get_project_report

GENERATE_REPORTS = False


def test_report(git_project):
    js_project = get_js_project()

    # Same created, later name
    gp2 = deepcopy(git_project)
    gp2.data["name"] = "zgitrepo"

    # Later created, earlier name
    gp3 = deepcopy(git_project)
    gp3.data["name"] = "agitrepo"
    eastern = pytz.timezone("US/Eastern")
    gp3.data["created"] = datetime.datetime(2020, 2, 1, 13, 0, 0, 0, tzinfo=eastern)

    report = get_project_report(add_projects=[gp2, gp3, git_project, js_project])

    attr_extensions = dict(
        json="json",
        yaml="yml",
        latex="tex",
    )

    for attr, extension in attr_extensions.items():
        report_outpath = os.path.join(REPORTS_FOLDER, f"report.{extension}")
        report_body = getattr(report, attr)
        if GENERATE_REPORTS:
            with open(report_outpath, "w") as f:
                f.write(report_body)
        else:
            with open(report_outpath, "r") as f:
                expected_report_body = f.read()
            assert report_body == expected_report_body



"""
A set of tools for describing software projects. Finds software projects, analyzes them,
and outputs reports.
"""
from projectreport.analyzer.project import Project
from projectreport.analyzer.ts.github import GithubAnalysis
from projectreport.config import DEFAULT_IGNORE_PATHS
from projectreport.finder.combine import CombinedFinder
from projectreport.finder.git import GitFinder
from projectreport.finder.js import JavaScriptPackageFinder
from projectreport.finder.python import PythonPackageFinder
from projectreport.report.report import Report



"""
Tools to search for regex within files and return lines around the match
"""
from projectreport.searcher.search import (
    read_all_files_in_folder_print_lines_around_regex,
    read_all_files_in_folders_print_lines_around_regex,
    read_file_get_lines_around_regex,
)

__all__ = [
    "read_all_files_in_folder_print_lines_around_regex",
    "read_all_files_in_folders_print_lines_around_regex",
    "read_file_get_lines_around_regex",
]



from projectreport.finder.base import Finder


class GitFinder(Finder):
    """
    Finder that automatically looks for projects with git repositories.
    """

    def __init__(self, recursive: bool = True):
        super().__init__(recursive=recursive, required_folders=(".git",))



from collections import UserList
from typing import Any, Iterable, Sequence


class RotatingList(UserList):
    """
    A list which will hold only max_len items. Once more items are added then
    can fit in the list, items will be removed.

    When appending or extending items which will make the list larger than
    max_len, the earliest items in the list will
    be removed. When inserting an item which will make the list larger than
    max_len, the last item will be removed.
    """

    def __init__(self, items: Sequence[Any], max_len: int):
        self.max_len = max_len
        super().__init__(items)

    def append(self, item: Any):
        if len(self) == self.max_len:
            self.pop(0)
        super().append(item)

    def insert(self, index: int, item: Any):
        if len(self) == self.max_len:
            self.pop(-1)
        super().insert(index, item)

    def extend(self, items: Iterable[Any]):
        # TODO [#4]: more efficient implementation of RotatingList.extend
        [self.append(item) for item in items]
        # if len(items) >= self.max_len:
        #     self.data = items[-self.max_len:]
        # elif len(self) + len(items) > self.max_len:



from pathlib import Path

from cached_property import cached_property


class Parser:
    def __init__(self, path: str):
        self.path = path

    @cached_property
    def parsed(self):
        raise NotImplementedError

    @cached_property
    def contents(self) -> str:
        with open(self.path, encoding="utf8") as f:
            file_contents = f.read()
        return file_contents

    @cached_property
    def docstring(self):
        raise NotImplementedError

    @cached_property
    def version(self):
        raise NotImplementedError

    @cached_property
    def file_name(self) -> str:
        return Path(self.path).with_suffix("").name



import ast
import importlib.util
import warnings
from types import ModuleType
from typing import Optional

from cached_property import cached_property
from getversion import get_module_version
from getversion.main import ModuleVersionNotFound, get_version_using_pkgresources
from getversion.plugin_builtins import get_builtin_module_version
from getversion.plugin_eggs_and_wheels import get_unzipped_wheel_or_egg_version
from getversion.plugin_importlib_metadata import get_version_using_importlib_metadata

from projectreport.analyzer.parsers.base import Parser
from projectreport.version import Version


class PythonParser(Parser):
    @cached_property
    def parsed(self):
        try:
            return ast.parse(self.contents)
        except SyntaxError:
            warnings.warn(f"Could not parse {self.path} due to SyntaxError")
            return None

    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parsed is None:
            return None
        return ast.get_docstring(self.parsed)

    @cached_property
    def module(self) -> ModuleType:
        spec = importlib.util.spec_from_file_location(self.file_name, self.path)
        module = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(module)  # type: ignore
        return module

    @cached_property
    def version(self) -> Optional[Version]:
        try:
            version, details = get_module_version(
                self.module,
                rootmodule_strategies=(
                    get_version_using_pkgresources,
                    get_builtin_module_version,
                    get_version_using_importlib_metadata,
                    get_unzipped_wheel_or_egg_version,
                    # get_version_using_setuptools_scm,  # was returning invalid results
                ),
            )
        except ModuleVersionNotFound:
            return None
        return Version.from_str(version)



from dataclasses import dataclass
from typing import Optional, Tuple

import pandas as pd
import semver


@dataclass
class Version:
    major: Optional[int] = None
    minor: Optional[int] = None
    patch: Optional[int] = None
    prerelease: Optional[str] = None
    build: Optional[str] = None
    full_str: Optional[str] = None
    is_semver: bool = False

    def __post_init__(self):
        self._validate()

    def _validate(self):
        attrs = [
            "major",
            "minor",
            "patch",
            "prerelease",
            "build",
            "full_str",
        ]
        if all([getattr(self, attr) is None for attr in attrs]):
            raise ValueError(
                "must provide some version information, all attributes are None"
            )

    @property
    def semantic_version(self) -> semver.VersionInfo:
        if not self.is_semver:
            raise ValueError("is_semver=False, cannot create semver object")

        return semver.VersionInfo(
            major=self.major,
            minor=self.minor,
            patch=self.patch,
            prerelease=self.prerelease,
            build=self.build,
        )

    def __eq__(self, other):
        if not self.is_semver or isinstance(other, Version) and not other.is_semver:
            return super().__eq__(other)
        if isinstance(other, Version):
            compare = other.semantic_version
        else:
            compare = other
        return self.semantic_version == compare

    def __lt__(self, other):
        if not self.is_semver:
            raise ValueError("cannot compare non-semantic versions")
        if isinstance(other, Version):
            compare = other.semantic_version
        else:
            compare = other
        return self.semantic_version < compare

    def __le__(self, other):
        if not self.is_semver:
            raise ValueError("cannot compare non-semantic versions")
        if isinstance(other, Version):
            compare = other.semantic_version
        else:
            compare = other
        return self.semantic_version <= compare

    def __gt__(self, other):
        if not self.is_semver:
            raise ValueError("cannot compare non-semantic versions")
        if isinstance(other, Version):
            compare = other.semantic_version
        else:
            compare = other
        return self.semantic_version > compare

    def __ge__(self, other):
        if not self.is_semver:
            raise ValueError("cannot compare non-semantic versions")
        if isinstance(other, Version):
            compare = other.semantic_version
        else:
            compare = other
        return self.semantic_version >= compare

    @classmethod
    def from_str(cls, version: str) -> "Version":
        full_str = version
        if version.startswith("v") and version[1].isdigit():
            # Looks like could be semantic versioning but starting with v, strip the v
            # but keep it in full_str
            version = version[1:]

        sem_version: Optional[semver.VersionInfo] = None
        try:
            sem_version = semver.VersionInfo.parse(version)
        except ValueError as e:
            if "not valid SemVer string" not in str(e):
                raise e
        if sem_version is not None:
            return cls.from_semver_version(sem_version, full_str=full_str)

        # Not semantic versioning, don't parse it
        return cls(full_str=full_str)

    @classmethod
    def from_semver_version(
        cls, version: semver.VersionInfo, full_str: Optional[str] = None
    ) -> "Version":
        return cls(
            major=version.major,
            minor=version.minor,
            patch=version.patch,
            prerelease=version.prerelease,
            build=version.build,
            full_str=full_str,
            is_semver=True,
        )

    def __str__(self) -> str:
        if self.full_str is not None:
            return self.full_str

        return str(self.semantic_version)


def major_minor_patch(
    version: Version,
) -> Tuple[Optional[int], Optional[int], Optional[int]]:
    if not version.is_semver:
        return None, None, None
    return version.major, version.minor, version.patch


def add_major_minor_patch_to_df(df: pd.DataFrame, version_col: str = "tag_name"):
    df["Version"] = df[version_col].apply(lambda tag: Version.from_str(tag))
    df["Major"], df["Minor"], df["Patch"] = zip(*df["Version"].map(major_minor_patch))


def add_major_minor_patch_changed_to_df(df: pd.DataFrame):
    df["Max Version"] = df["Version"].cummax()
    df["Max Major"], df["Max Minor"], df["Max Patch"] = zip(
        *df["Max Version"].map(major_minor_patch)
    )
    for col in ["Major", "Minor", "Patch"]:
        df[f"Max {col}"] = df[f"Max {col}"].shift()
        df[f"{col} Diff"] = df[col] - df[f"Max {col}"]
    df["Major Changed"] = False
    df.loc[
        (~pd.isnull(df["Major Diff"])) & (df["Major Diff"] != 0), "Major Changed"
    ] = True
    df["Minor Changed"] = False
    df.loc[
        (~pd.isnull(df["Minor Diff"]))
        & (df["Minor Diff"] != 0)
        & (~df["Major Changed"]),
        "Minor Changed",
    ] = True
    df["Patch Changed"] = False
    df.loc[
        (~pd.isnull(df["Patch Diff"]))
        & (df["Patch Diff"] != 0)
        & (~df["Minor Changed"])
        & (~df["Major Changed"]),
        "Patch Changed",
    ] = True



import warnings
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Callable, Dict, List, Optional, Union

import dateutil
import pandas as pd
from dateutil import parser as dateparser
from github.Commit import Commit
from github.CommitStats import CommitStats
from github.GitAuthor import GitAuthor
from github.GitCommit import GitCommit
from github.GitRelease import GitRelease
from github.Issue import Issue
from github.NamedUser import NamedUser
from github.Repository import Repository
from github.Stargazer import Stargazer

from projectreport.analyzer.ts.base import TimeSeriesAnalysis
from projectreport.analyzer.ts.types import DictList
from projectreport.tools.monkey_patch_github import (
    NoMorePagesAllowedException,
    monkey_patch_github_obj_for_throttling,
)
from projectreport.version import (
    add_major_minor_patch_changed_to_df,
    add_major_minor_patch_to_df,
)


class GithubAnalysis(TimeSeriesAnalysis):
    analysis_attrs = ["repo"]

    def __init__(self, repo: Repository, auto_throttle: bool = True):
        self.repo = deepcopy(repo)
        self.auto_throttle = auto_throttle
        if self.auto_throttle:
            monkey_patch_github_obj_for_throttling(self.repo)

    @property
    def event_functions(self) -> Dict[str, Callable[[Repository], DictList]]:
        funcs: Dict[str, Callable[[Repository], DictList]] = dict(
            commits=commit_stats_from_repo,
            issues=issue_stats_from_repo,
            stars=stars_from_repo,
            releases=releases_from_repo,
        )
        return funcs

    @property
    def count_functions(self) -> Dict[str, Callable[[DictList, str], DictList]]:
        funcs: Dict[str, Callable[[DictList, str], DictList]] = dict(
            commits=commit_loc_counts_from_commit_events,
            issues=issue_counts_from_issue_events,
            stars=star_counts_from_star_events,
            releases=release_counts_from_release_events,
        )
        return funcs


def commit_stats_from_repo(repo: Repository, author_stats: bool = True) -> DictList:
    all_data = []
    commit: Commit
    try:
        for commit in repo.get_commits():
            stats: CommitStats = commit.stats
            author: Optional[Union[NamedUser, GitAuthor]] = _get_author_from_commit(
                commit
            )
            committer: Optional[
                Union[NamedUser, GitAuthor]
            ] = _get_committer_from_commit(commit)
            data_dict = dict(
                sha=commit.sha,
                last_modified=dateparser.parse(commit.last_modified)
                if commit.last_modified is not None
                else None,
                additions=stats.additions,
                deletions=stats.deletions,
                url=commit.html_url,
            )
            if author_stats:
                if author is not None:
                    data_dict.update(_get_data_from_named_user_or_git_author(author))
                if committer is not None:
                    data_dict.update(
                        _get_data_from_named_user_or_git_author(
                            committer, is_committer=True
                        )
                    )
            all_data.append(data_dict)
    except NoMorePagesAllowedException:
        warnings.warn(
            f"Could not collect full history for {repo.name} commits as Github "
            f"limits the amount of history than can be pulled"
        )

    return all_data  # type: ignore


def commit_loc_counts_from_commit_events(
    commits: DictList, freq: str = "d"
) -> DictList:
    event_df = pd.DataFrame(commits)
    event_df["net"] = event_df["additions"] - event_df["deletions"]
    event_df["change"] = event_df["additions"] + event_df["deletions"]
    start = _get_end_of_period(event_df["last_modified"].min(), freq)
    end = event_df["last_modified"].max()

    dates = pd.date_range(start=start, end=end, freq=freq)
    count_data = []
    for date in dates:
        until_time_df = event_df[event_df["last_modified"] < date]
        commit_counts = len(until_time_df)
        loc = until_time_df["net"].sum()
        loc_changed = until_time_df["change"].sum()
        count_data.append(
            dict(date=date, commits=commit_counts, loc=loc, loc_changed=loc_changed)
        )
    return count_data


def issue_stats_from_repo(repo: Repository) -> DictList:
    all_data = []
    issue: Issue
    try:
        for issue in repo.get_issues(state="all"):
            data_dict = dict(
                number=issue.number,
                created_at=issue.created_at,
                updated_at=issue.updated_at,
                closed_at=issue.closed_at,
                comments_count=issue.comments,
                state=issue.state,
                is_pull_issue=issue.pull_request is not None,
            )
            all_data.append(data_dict)
    except NoMorePagesAllowedException:
        warnings.warn(
            f"Could not collect full history for {repo.name} issues as Github "
            f"limits the amount of history than can be pulled"
        )

    return all_data  # type: ignore


def issue_counts_from_issue_events(issues: DictList, freq: str = "d") -> DictList:
    event_df = pd.DataFrame(issues)
    start = _get_end_of_period(event_df["created_at"].min(), freq)
    end = event_df["updated_at"].max()
    dates = pd.date_range(start=start, end=end, freq=freq)
    count_data = []
    for date in dates:
        until_time_df = event_df.loc[event_df["created_at"] <= date]
        # Mark issues which are closed now as open if they were not closed by this time
        until_time_df.loc[until_time_df["closed_at"] > date, "state"] = "open"

        pull_df = until_time_df[until_time_df["is_pull_issue"]]
        issue_df = until_time_df[~until_time_df["is_pull_issue"]]

        all_issues = len(until_time_df)
        closed_issues = len(issue_df[issue_df["state"] == "closed"])
        closed_pull_issues = len(pull_df[pull_df["state"] == "closed"])
        open_issues = len(issue_df[issue_df["state"] == "open"])
        open_pull_issues = len(pull_df[pull_df["state"] == "open"])

        count_data.append(
            dict(
                date=date,
                all_issues=all_issues,
                closed_issues=closed_issues,
                closed_pull_issues=closed_pull_issues,
                open_issues=open_issues,
                open_pull_issues=open_pull_issues,
            )
        )

    return count_data


def stars_from_repo(repo: Repository) -> DictList:
    all_data = []
    stars: Stargazer
    try:
        for stars in repo.get_stargazers_with_dates():
            user: NamedUser = stars.user
            data_dict = dict(
                date=stars.starred_at,
                user_name=user.name,
                user_login=user.login,
            )
            all_data.append(data_dict)
    except NoMorePagesAllowedException:
        warnings.warn(
            f"Could not collect full history for {repo.name} stars as Github "
            f"limits the amount of history than can be pulled"
        )

    return all_data  # type: ignore


def star_counts_from_star_events(stars: DictList, freq: str = "d") -> DictList:
    event_df = pd.DataFrame(stars)
    start = _get_end_of_period(event_df["date"].min(), freq)
    end = event_df["date"].max()

    dates = pd.date_range(start=start, end=end, freq=freq)
    count_data = []
    for date in dates:
        until_time_df = event_df[event_df["date"] < date]
        star_count = len(until_time_df)
        count_data.append(dict(date=date, stars=star_count))

    return count_data


def releases_from_repo(repo: Repository) -> DictList:
    all_data = []
    releases: GitRelease
    try:
        for release in repo.get_releases():
            author = release.author
            author_name: Optional[str] = None
            author_login: Optional[str] = None
            if author is not None:
                author_name = author.name
                author_login = author.login
            data_dict = dict(
                created_at=release.created_at,
                published_at=release.published_at,
                url=release.html_url,
                id=release.id,
                tag_name=release.tag_name,
                draft=release.draft,
                prerelease=release.prerelease,
                body=release.body,
                author_name=author_name,
                author_login=author_login,
            )
            all_data.append(data_dict)
    except NoMorePagesAllowedException:
        warnings.warn(
            f"Could not collect full history for {repo.name} releases as Github "
            f"limits the amount of history than can be pulled"
        )

    return all_data  # type: ignore


def release_counts_from_release_events(
    releases: DictList,
    freq: str = "d",
    date_var: str = "published_at",
) -> DictList:
    event_df = pd.DataFrame(releases)
    event_df.sort_values(date_var, inplace=True)
    start = _get_end_of_period(event_df[date_var].min(), freq)
    end = event_df[date_var].max()
    add_major_minor_patch_to_df(event_df)

    semver_df = event_df.loc[event_df["Version"].apply(lambda ver: ver.is_semver)]
    if not semver_df.empty:
        add_major_minor_patch_changed_to_df(semver_df)

    dates = pd.date_range(start=start, end=end, freq=freq)
    count_data = []
    for date in dates:
        until_time_df = event_df[event_df[date_var] < date]
        release_count = len(until_time_df)
        until_time_df = semver_df[semver_df[date_var] < date]
        if not until_time_df.empty:
            major_df = until_time_df[until_time_df["Major Changed"]]
            major_count = len(major_df)
            minor_df = until_time_df[until_time_df["Minor Changed"]]
            minor_count = len(minor_df)
            patch_df = until_time_df[until_time_df["Patch Changed"]]
            patch_count = len(patch_df)
        else:
            major_count, minor_count, patch_count = 0, 0, 0
        count_data.append(
            dict(
                date=date,
                releases=release_count,
                major_releases=major_count,
                minor_releases=minor_count,
                patch_releases=patch_count,
            )
        )

    return count_data


def _get_data_from_named_user_or_git_author(
    user: Union[NamedUser, GitAuthor], is_committer: bool = False
) -> Dict[str, str]:
    if is_committer:
        key_base = "committer"
    else:
        key_base = "author"

    data: Dict[str, str] = {
        f"{key_base}_name": user.name,
        f"{key_base}_email": user.email,
    }

    if isinstance(user, NamedUser):
        data.update(
            {
                f"{key_base}_login": user.login,
            }
        )

    return data


def _get_author_from_commit(commit: Commit) -> Optional[Union[NamedUser, GitAuthor]]:
    if commit.author is not None:
        # NamedUser
        return commit.author

    git_commit: GitCommit = commit.commit
    # GitAuthor
    return git_commit.author


def _get_committer_from_commit(commit: Commit) -> Optional[Union[NamedUser, GitAuthor]]:
    if commit.committer is not None:
        # NamedUser
        return commit.committer

    git_commit: GitCommit = commit.commit
    # GitAuthor
    return git_commit.committer


def _get_end_of_period(date: pd.Timestamp, freq: str) -> pd.Timestamp:
    # TODO [#16]: get _get_end_of_period working correctly for all frequencies
    #
    # Works correctly for month, day, hour, and weeks starting on a different day.
    # Currently gets beginning of period for weeks starting with the same day.
    try:
        return date.ceil(freq)
    except ValueError as e:
        if "is a non-fixed frequency" in str(e):
            return date.to_period(freq).to_timestamp(freq).tz_localize("UTC")
        else:
            raise e



import time
import types
from datetime import datetime, timedelta, timezone
from typing import List, Optional, Union

from github import Github, GithubException, RateLimitExceededException
from github.GithubObject import GithubObject
from github.Rate import Rate
from github.RateLimit import RateLimit
from github.Requester import Requester
from requests.exceptions import RequestException

from projectreport.logger import logger

RATE_LIMIT_PAD_SECONDS = 180
REQUEST_EXCEPTION_WAIT = 60


def monkey_patch_github_obj_for_throttling(gh_obj: Union[Github, GithubObject]):
    requester = _get_requester(gh_obj)
    orig_request_and_check = requester.requestJsonAndCheck

    def request_json_and_check_patched_for_throttling(
        obj: Requester,
        verb: str,
        url: str,
        retries_remaining: int = 3,
        collected_exceptions: Optional[List[Exception]] = None,
        **kwargs,
    ):
        if retries_remaining <= 0:
            raise GithubException(
                f"could not make github request after retries. Got errors {collected_exceptions}",
                {"exc": collected_exceptions},
                None,
            )
        if collected_exceptions is None:
            collected_exceptions = []

        logger.debug(f"made {verb} request to {url}")
        try:
            return orig_request_and_check(verb, url, **kwargs)
        except RateLimitExceededException as e:
            collected_exceptions.append(e)
            _get_rate_limit_and_wait(requester)
            return request_json_and_check_patched_for_throttling(
                obj,
                verb,
                url,
                retries_remaining=retries_remaining - 1,
                collected_exceptions=collected_exceptions,
                **kwargs,
            )
        except (RequestException, GithubException) as e:
            if "pagination is limited for this resource" in str(e).casefold():
                raise NoMorePagesAllowedException
            collected_exceptions.append(e)
            _wait_for_read_timeout(e)
            return request_json_and_check_patched_for_throttling(
                obj,
                verb,
                url,
                retries_remaining=retries_remaining - 1,
                collected_exceptions=collected_exceptions,
                **kwargs,
            )

    requester.requestJsonAndCheck = types.MethodType(  # type: ignore
        request_json_and_check_patched_for_throttling, requester
    )


def _wait_for_read_timeout(e: Union[RequestException, GithubException]):
    logger.warning(
        f"Got request/Github exception: {e}. Sleeping for {REQUEST_EXCEPTION_WAIT}s"
    )
    time.sleep(REQUEST_EXCEPTION_WAIT)
    logger.info(f"Resuming from read timeout sleep at {datetime.now()}")


def _get_rate_limit_and_wait(obj: Requester):
    headers, data = obj.requestJsonAndCheck("GET", "/rate_limit")
    if data is None:
        raise ValueError("got no data from request")
    limits = RateLimit(obj, headers, data["resources"], True)
    reset = _get_last_reset_datetime_utc(limits)
    now = datetime.now(timezone.utc)
    diff = reset - now
    seconds = diff.total_seconds()
    current_tz = datetime.utcnow().astimezone().tzinfo
    logger.info(f"Rate limit exceeded")
    logger.info(f"Reset is at {reset.astimezone(current_tz)}.")
    if seconds > 0.0:
        # Extra 10s to ensure it's ready
        logger.info(f"Waiting for {diff + timedelta(seconds=RATE_LIMIT_PAD_SECONDS)}")
        time.sleep(seconds + RATE_LIMIT_PAD_SECONDS)
        logger.info(f"Done waiting - resuming at {datetime.now()}")


def _get_last_reset_datetime_utc(limit: RateLimit) -> datetime:
    # For some reason integration_manifest and source_import are in the raw data
    # but not supported by RateLimit object
    limit_attrs = [
        "core",
        "search",
        "graphql",
        # 'integration_manifest',
        # 'source_import',
        "rate",
    ]
    all_resets = []
    for attr in limit_attrs:
        try:
            rate: Rate = getattr(limit, attr)
        except AttributeError:
            continue
        reset = rate.reset.replace(tzinfo=timezone.utc)
        all_resets.append(reset)
    if not all_resets:
        raise ValueError(
            f"no reset time in rate limit object: {limit} with data {limit.raw_data}"
        )

    last_reset = max(all_resets)
    return last_reset


def _get_requester(gh_obj: Union[Github, GithubObject]) -> Requester:
    if isinstance(gh_obj, Github):
        attr = "_Github__requester"
    else:
        attr = "_requester"
    requester: Requester = getattr(gh_obj, attr)
    return requester


class NoMorePagesAllowedException(Exception):
    pass



import datetime
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Sequence, Union

from cached_property import cached_property

if TYPE_CHECKING:
    from projectreport.analyzer.project import Project

from projectreport.report.base import BaseReport
from projectreport.report.project import ProjectReport


class Report(BaseReport):
    """
    Pass Projects to create a report of the analysis of the projects.
    """

    def __init__(self, projects: Sequence["Project"], depth: int = 0):
        self.project_reports = [
            ProjectReport(project, depth=depth) for project in projects
        ]

    @cached_property
    def data(self) -> List[dict]:
        items = [report.data for report in self.project_reports]
        items.sort(key=_created_then_name_key)
        return items

    @cached_property
    def doc(self):
        from projectreport.report.latex import multi_project_latex_document

        return multi_project_latex_document(self.data)

    def sort(self, key: Optional[Callable[[dict], str]] = None, reverse: bool = False):
        self.data.sort(key=key, reverse=reverse)

    def default_sort(self):
        self.data.sort(key=_created_then_name_key)


def _created_then_name_key(data: dict) -> str:
    created: Optional[datetime.datetime] = data["created"]
    if created is None:
        created_str = "0000-00-00 00:00:00.000000"
    else:
        created_str = created.strftime("%Y-%m-%d %H:%M:%S.%f")

    name: str = data["name"]
    return f"{created_str}_{name}"



import os
from typing import Sequence

from projectreport.finder.base import Finder


class CombinedFinder(Finder):
    """
    Finder which can be passed other specific finders, and will return project paths that match any of the
    passed finders.
    """

    def __init__(self, finders: Sequence[Finder], recursive: bool = True):
        self.finders = finders
        super().__init__(recursive=recursive)
        self.file_extensions = []
        for finder in self.finders:
            if finder.file_extensions:
                self.file_extensions.extend(finder.file_extensions)

    def is_valid(self, path: str):
        is_valid = False
        for finder in self.finders:
            if finder.is_valid(path):
                is_valid = True
        return is_valid



import contextlib
import shutil
import tempfile
from pathlib import Path
from typing import Iterator


@contextlib.contextmanager
def create_temp_path() -> Iterator[Path]:
    """
    Returns a temporary folder path

    Use this instead of tempfile.TemporaryDirectory because:
    1. That returns a string and not a path
    2. On MacOS, the temp directory has a symlink. This resolves the symlink so that
       there won't be any mismatch in resolved paths.
    3. On Windows, the temp directory can fail to delete with a PermissionError. This function
       will try to delete the temp directory, but if it fails with an error it will just ignore it.
    """
    temp_dir = tempfile.TemporaryDirectory()
    temp_path = Path(temp_dir.name).resolve()
    yield temp_path
    shutil.rmtree(temp_path, ignore_errors=True)



from pathlib import Path

GITHUB_REPO_URL = "https://github.com/nickderobertis/github-project-example"
TESTS_DIR = Path(__file__).parent
PROJECT_DIR = TESTS_DIR.parent
TEST_FILES_BASE_PATH = TESTS_DIR / "input_data"
PYTHON_PROJECT_NAME = "python_example"
PYTHON_PROJECT_PATH = TEST_FILES_BASE_PATH / PYTHON_PROJECT_NAME
JS_PROJECT_NAME = "js_example"
JS_PROJECT_PATH = TEST_FILES_BASE_PATH / JS_PROJECT_NAME
GIT_PROJECT_NAME = "git_example"
REPORTS_NAME = "reports"
REPORTS_FOLDER = TEST_FILES_BASE_PATH / REPORTS_NAME



from projectreport import JavaScriptPackageFinder
from projectreport.finder.combine import CombinedFinder
from projectreport.finder.git import GitFinder
from projectreport.finder.python import PythonPackageFinder
from tests.config import (
    JS_PROJECT_PATH,
    PROJECT_DIR,
    PYTHON_PROJECT_PATH,
    TEST_FILES_BASE_PATH,
)


def test_find_python_package():
    finder = PythonPackageFinder()
    search_paths = [TEST_FILES_BASE_PATH]
    project_paths = finder.find_all(search_paths)
    assert project_paths == [str(PYTHON_PROJECT_PATH)]


def test_find_javascript_package():
    finder = JavaScriptPackageFinder()
    search_paths = [TEST_FILES_BASE_PATH]
    project_paths = finder.find_all(search_paths)
    assert project_paths == [str(JS_PROJECT_PATH)]


def test_find_git_package():
    finder = GitFinder()
    search_paths = [PROJECT_DIR]
    project_paths = finder.find_all(search_paths)
    assert project_paths == [PROJECT_DIR]


def test_combined_finder():
    finder = CombinedFinder(
        [
            GitFinder(),
            PythonPackageFinder(),
            JavaScriptPackageFinder(),
        ]
    )
    search_paths = [TEST_FILES_BASE_PATH]
    project_paths = finder.find_all(search_paths)
    assert sorted(project_paths) == [str(JS_PROJECT_PATH), str(PYTHON_PROJECT_PATH)]



import yaml
from cached_property import cached_property

from projectreport.report.json import to_json


class BaseReport:
    def __str__(self):
        return self.yaml

    @cached_property
    def json(self) -> str:
        return to_json(self.data)

    @cached_property
    def yaml(self) -> str:
        return yaml.dump(self.data, indent=2)

    @cached_property
    def latex(self):
        return str(self.doc)

    #### Scaffolding below, implement these in subclass ########

    @cached_property
    def data(self):
        raise NotImplementedError

    @cached_property
    def doc(self):
        raise NotImplementedError



from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from projectreport.analyzer.project import Project
    from projectreport.analyzer.parsers.base import Parser

import os

from cached_property import cached_property

from projectreport.analyzer.analysis import ModuleAnalysis
from projectreport.analyzer.analyzable import Analyzable
from projectreport.analyzer.parsers.index import PARSER_EXTENSIONS


class Module(Analyzable):
    def __init__(
        self,
        path: str,
        package: Optional[str] = None,
        project: Optional["Project"] = None,
    ):
        base_path, extension = os.path.splitext(path)
        self.name = os.path.basename(base_path)
        self.extension = extension.strip(".")
        if package is None:
            package = self.name
        self.package = package
        super().__init__(path, project=project)
        self.analysis = ModuleAnalysis(self)

    @cached_property
    def parser(self) -> Optional["Parser"]:
        if self.extension not in PARSER_EXTENSIONS:
            return None
        return PARSER_EXTENSIONS[self.extension](self.path)



from typing import TYPE_CHECKING, Dict, Optional, Union

from projectreport.data import AnalysisData
from projectreport.version import Version

if TYPE_CHECKING:
    from projectreport.analyzer.project import Project
    from projectreport.analyzer.analysis import Analysis
    from projectreport.analyzer.parsers.base import Parser

from cached_property import cached_property


class Analyzable:
    parser: Optional["Parser"] = None
    analysis: Optional["Analysis"] = None
    name: Optional[str] = None

    def __init__(self, path: str, project: Optional["Project"] = None):
        self.path = path
        self.project = project

    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parser is None:
            return None
        return self.parser.docstring

    @cached_property
    def version(self) -> Optional[Version]:
        if self.parser is None:
            return None
        return self.parser.version

    @cached_property
    def data(self) -> AnalysisData:
        if self.analysis is None:
            raise ValueError(
                "cannot get data from Analyzable if no analysis is attached"
            )

        data: AnalysisData = {}
        data.update(self.analysis.data)
        data.update(
            dict(
                docstring=self.docstring,
                version=str(self.version) if self.version is not None else None,
            )
        )
        if self.name is not None:
            data.update(dict(name=self.name))
        return data



from datetime import datetime
from typing import Any, Callable, Dict, List, Sequence, Union

from github.Repository import Repository

from projectreport.analyzer.ts.types import DictList


class TimeSeriesAnalysis:
    analysis_attrs: Sequence[str] = tuple()

    def get_event_data(self, item: str, **kwargs) -> DictList:
        func = self.event_functions[item]
        return func(*self.analysis_items, **kwargs)

    def get_counts(self, item: str, freq: str, **kwargs) -> DictList:
        event_data = self.get_event_data(item)
        func = self.count_functions[item]
        return func(event_data, freq, **kwargs)

    @property
    def event_functions(self) -> Dict[str, Callable]:
        raise NotImplementedError

    @property
    def count_functions(self) -> Dict[str, Callable]:
        raise NotImplementedError

    @property
    def analysis_items(self) -> List[Any]:
        return [getattr(self, attr) for attr in self.analysis_attrs]

    @property
    def supported_items(self) -> List[str]:
        return list(self.event_functions.keys())



from typing import List

import pyexlatex as pl
from pyexlatex.models.format.breaks import OutputLineBreak
from pyexlatex.models.item import ItemBase
from pyexlatex.models.section.base import TextAreaMixin


class SubProjectLatex(TextAreaMixin, ItemBase):
    def __init__(self, data, **kwargs):
        contents = self.get_contents(data)
        super().__init__(None, contents, **kwargs)

    def get_contents(self, data):
        line_break = OutputLineBreak()
        green = pl.RGB(42, 138, 11, color_name="darkgreen")
        commits_str = (
            "Commits" if data["num_commits"] and data["num_commits"] > 1 else "Commit"
        )
        contents = [
            pl.Bold(data["name"]) if data["name"] else None,
            pl.TextColor(f"{data['lines']} LOC", color=green)
            if data["lines"]
            else None,
            ["|", pl.TextColor(f"{data['num_commits']} {commits_str}", color="blue")]
            if data["num_commits"]
            else None,
            ["|", f'Created {data["created"].date()}'] if data["created"] else None,
            ["|", f'Updated {data["updated"].date()}'] if data["updated"] else None,
            line_break,
            data["docstring"],
        ]
        contents = [content for content in contents if content is not None]
        if not contents[-1] == line_break:
            contents.append(line_break)
        return contents

    def __str__(self):
        from pyexlatex.logic.builder import _build

        if isinstance(self.contents, str):
            return self.contents
        return _build(self.contents)


def project_latex(data: dict):
    contents = []
    contents.append(SubProjectLatex(data))
    if "subprojects" in data:
        for project in data["subprojects"]:
            contents.append(
                pl.ParagraphIndent(project_latex(data["subprojects"][project]))
            )
    return contents


def multi_project_latex(data: List[dict]):
    contents = []
    for data_dict in data:
        contents.extend(
            [
                project_latex(data_dict),
                pl.HLine(),
                pl.VSpace(0.5),
            ]
        )

    del contents[-2:]  # remove last horizontal line and vertical spacing
    return contents


def project_latex_document(data) -> pl.Document:
    contents = project_latex(data)
    return _get_document(contents)


def multi_project_latex_document(data: List[dict]) -> pl.Document:
    contents = multi_project_latex(data)
    return _get_document(contents)


def _get_document(contents) -> pl.Document:
    doc = pl.Document(contents, title="Project Report")
    return doc



import datetime
import json


def to_json(dict_: dict) -> str:
    def json_converter(o):
        if isinstance(o, datetime.datetime):
            return o.__str__()

    return json.dumps(dict_, default=json_converter, indent=2)



import os
from pathlib import Path
from typing import Optional, Sequence, Union

import git
from cached_property import cached_property

from projectreport.analyzer.folder import Folder
from projectreport.config import DEFAULT_IGNORE_PATHS


class Project(Folder):
    """
    Pass a project path to get analysis about the project.
    """

    def __init__(
        self,
        path: Union[str, Path],
        excluded_types: Optional[Sequence[str]] = None,
        included_types: Optional[Sequence[str]] = None,
        ignore_paths: Optional[Sequence[str]] = DEFAULT_IGNORE_PATHS,
    ):
        super().__init__(
            path,
            project=self,
            excluded_types=excluded_types,
            included_types=included_types,
            ignore_paths=ignore_paths,
        )

    @cached_property
    def repo(self) -> Optional[git.Repo]:
        try:
            return git.Repo(self.path)
        except git.InvalidGitRepositoryError:
            return None



from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Sequence, Union

from projectreport.logger import logger

if TYPE_CHECKING:
    from projectreport.analyzer.project import Project
    from projectreport.analyzer.parsers.base import Parser

import os

from cached_property import cached_property

from projectreport.analyzer.analysis import FolderAnalysis
from projectreport.analyzer.analyzable import Analyzable
from projectreport.analyzer.module import Module
from projectreport.analyzer.parsers.index import PARSER_DOC_FILES
from projectreport.config import DEFAULT_IGNORE_PATHS
from projectreport.tools.expand_glob import all_possible_paths


class Folder(Analyzable):
    def __init__(
        self,
        path: Union[str, Path],
        project: Optional["Project"] = None,
        excluded_types: Optional[Sequence[str]] = None,
        included_types: Optional[Sequence[str]] = None,
        ignore_paths: Optional[Sequence[str]] = DEFAULT_IGNORE_PATHS,
    ):
        from projectreport.analyzer.project import Project

        path = os.path.abspath(path)
        self.name = os.path.basename(path)
        self.options = dict(
            excluded_types=excluded_types,
            included_types=included_types,
            ignore_paths=ignore_paths,
        )
        if project is None:
            project = Project(self.path, **self.options)
        self._validate()
        super().__init__(path, project=project)

    @cached_property
    def file_names(self) -> List[str]:
        file_names = [file for file in next(os.walk(self.path))[2]]
        return file_names

    @cached_property
    def file_paths(self) -> List[str]:
        included_files = []
        if self.options["ignore_paths"] is not None:
            all_ignore_paths = all_possible_paths(
                self.options["ignore_paths"], self.path
            )

        for file in self.file_names:
            if self.options["ignore_paths"]:
                file_path = os.path.join(self.path, file)
                if file_path in all_ignore_paths:
                    continue  # this file excluded, skip it
            extension = os.path.splitext(file)[1].strip(".")
            if self.options["included_types"] is not None:
                if extension in self.options["included_types"]:
                    included_files.append(file)
            elif self.options["excluded_types"] is not None:
                if extension not in self.options["excluded_types"]:
                    included_files.append(file)
            else:
                # No included or excluded files, just keep all
                included_files.append(file)
        files = [os.path.join(self.path, name) for name in included_files]
        return files

    @cached_property
    def folder_paths(self) -> List[str]:
        folders = [file for file in next(os.walk(self.path))[1]]

        if self.options["ignore_paths"] is not None:
            all_ignore_paths = all_possible_paths(
                self.options["ignore_paths"], self.path
            )

        out_folders = []
        for folder in folders:
            abs_folder = os.path.join(self.path, folder)
            if self.options["ignore_paths"] is not None:
                if abs_folder in all_ignore_paths:
                    continue  # this folder excluded, skip it
            out_folders.append(abs_folder)

        return out_folders

    @cached_property
    def modules(self) -> List["Module"]:
        return [
            Module(file_path, self.name, project=self.project)
            for file_path in self.file_paths
        ]

    @cached_property
    def folders(self) -> List["Folder"]:
        all_folders = [
            Folder(folder, project=self.project, **self.options)
            for folder in self.folder_paths
        ]
        non_empty_folders = [folder for folder in all_folders if not folder.is_empty]
        return non_empty_folders

    @cached_property
    def analysis(self) -> "FolderAnalysis":
        analysis = FolderAnalysis(self)
        for folder in self.folders:
            logger.debug(f"Analyzing folder {folder.name}")
            analysis.add_subfolder_analysis(folder.analysis)
        for module in self.modules:
            logger.debug(f"Analyzing module {module.name}")
            analysis.add_module_analysis(module.analysis)
        return analysis

    @cached_property
    def parser(self) -> Optional["Parser"]:
        for file, parser in PARSER_DOC_FILES.items():
            if file in self.file_names:
                full_path = os.path.join(self.path, file)
                return PARSER_DOC_FILES[file](full_path)
        return None

    @cached_property
    def is_empty(self) -> bool:
        len_contents = len(self.file_paths) + len(self.folders)
        return len_contents == 0

    def _validate(self):
        if self.options["included_types"] and self.options["excluded_types"]:
            raise ValueError("cannot pass both include and exclude types")



import datetime
from pathlib import Path

import git
import pytest
import pytz

from projectreport import Project
from tests.config import GITHUB_REPO_URL, JS_PROJECT_PATH, PYTHON_PROJECT_PATH
from tests.dirutils import create_temp_path


@pytest.fixture(scope="session")
def github_project() -> Project:
    with create_temp_path() as temp_folder:
        project_path = temp_folder / "github-project-example"
        git.Repo.clone_from(GITHUB_REPO_URL, project_path)
        included_types = None
        project = Project(project_path, included_types=included_types)
        yield project


def get_python_project() -> Project:
    project_path = PYTHON_PROJECT_PATH
    included_types = ("py",)
    project = Project(project_path, included_types=included_types)
    return project


def get_js_project() -> Project:
    project_path = JS_PROJECT_PATH
    included_types = ("js",)
    project = Project(project_path, included_types=included_types)
    return project


@pytest.fixture(scope="session")
def git_project() -> Project:
    with create_temp_path() as base_dir:
        d = base_dir / "gitrepo"
        r = git.Repo.init(d)
        # This function just creates an empty file ...
        (d / "tempfile").write_text("woo")
        r.index.add(["tempfile"])
        r.index.commit("initial commit")
        included_types = None
        project = Project(d, included_types=included_types)
        eastern = pytz.timezone("US/Eastern")
        project.data["created"] = datetime.datetime(
            2020, 2, 1, 12, 0, 0, 0, tzinfo=eastern
        )
        project.data["updated"] = datetime.datetime(
            2020, 2, 1, 12, 1, 0, 0, tzinfo=eastern
        )
        assert project.path == str(d)
        yield project



import glob
import os
from typing import List, Optional, Sequence


def all_possible_paths(
    paths: Sequence[str], base_path: Optional[str] = None
) -> List[str]:
    all_paths = []
    for path in paths:
        # If got a relative path and a base path was passed
        if base_path is not None and not os.path.isabs(path):
            full_path = os.path.join(base_path, path)
        else:
            full_path = path
        all_paths.extend(glob.glob(full_path, recursive=False))
    return all_paths



import os
from typing import List, Optional, Sequence

from projectreport.config import DEFAULT_IGNORE_PATHS
from projectreport.tools.expand_glob import all_possible_paths


class Finder:
    """
    Class for finding projects by required folders or files, or looking for certain file extensions.
    """

    def __init__(
        self,
        recursive: bool = True,
        required_folders: Optional[Sequence[str]] = None,
        required_files: Optional[Sequence[str]] = None,
        file_extensions: Optional[Sequence[str]] = None,
    ):
        self.recursive = recursive
        self.required_folders = required_folders
        self.required_files = required_files
        self.file_extensions = file_extensions
        self.project_paths: List[str] = []

    def find(
        self, path: str, ignore_paths: Optional[Sequence[str]] = DEFAULT_IGNORE_PATHS
    ):
        if ignore_paths:
            all_ignore_paths = all_possible_paths(ignore_paths, path)
        else:
            all_ignore_paths = []

        if path in all_ignore_paths:
            # Ignored, do nothing and return
            return self.project_paths

        _, folders, files = next(os.walk(path))

        if self.is_valid(path):
            self.project_paths.append(path)
            if not self.recursive:
                # Stop searching a folder once a project is found
                return self.project_paths

        for folder in folders:
            folder_path = os.path.join(path, folder)
            if folder_path in all_ignore_paths:
                continue
            self.find(folder_path, ignore_paths=ignore_paths)

        return self.project_paths

    def find_all(
        self,
        paths: Sequence[str],
        ignore_paths: Optional[Sequence[str]] = DEFAULT_IGNORE_PATHS,
    ):
        [self.find(path, ignore_paths=ignore_paths) for path in paths]
        return self.project_paths

    def is_valid(self, path: str):
        _, folders, files = next(os.walk(path))
        if self.required_folders:
            for req_folder in self.required_folders:
                if req_folder not in folders:
                    return False
        if self.required_files:
            for req_file in self.required_files:
                if req_file not in files:
                    return False
        if self.file_extensions:
            extensions = {os.path.splitext(file)[1].strip(".") for file in files}
            has_extension = False
            for file_ext in self.file_extensions:
                if file_ext in extensions:
                    has_extension = True
            if not has_extension:
                return False

        return True

    def _validate(self):
        if not any([self.required_folders, self.required_files, self.file_extensions]):
            raise ValueError("must provide some conditions")



import json
import warnings
from typing import Any, Dict, Optional

from cached_property import cached_property

from projectreport.analyzer.parsers.base import Parser
from projectreport.version import Version


class PackageJSONParser(Parser):
    @cached_property
    def parsed(self) -> Optional[Dict[str, Any]]:
        try:
            return json.loads(self.contents)
        except SyntaxError:
            warnings.warn(f"Could not parse {self.path} due to SyntaxError")
            return None

    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parsed is None:
            return None
        return self.parsed.get("description")

    @cached_property
    def version(self) -> Optional[Version]:
        if self.parsed is None:
            return None
        return Version.from_str(self.parsed.get("version"))



import datetime
import os

import pytz
from git import Repo

from projectreport import Project
from projectreport.analyzer.analysis import FolderAnalysis
from projectreport.version import Version
from tests.config import (
    JS_PROJECT_NAME,
    JS_PROJECT_PATH,
    PYTHON_PROJECT_NAME,
    PYTHON_PROJECT_PATH,
)
from tests.fixtures.project import (
    get_js_project,
    get_python_project,
    git_project,
    github_project,
)


def test_python_project():
    project = get_python_project()

    assert project.file_names == ["__init__.py"]
    assert project.file_paths == [
        os.path.abspath(os.path.join(PYTHON_PROJECT_PATH, "__init__.py"))
    ]
    assert project.docstring == "An example Python package for testing purposes"
    assert project.version == Version.from_str("0.0.1")
    assert project.name == PYTHON_PROJECT_NAME
    assert project.path == os.path.abspath(PYTHON_PROJECT_PATH)
    assert project.repo is None
    assert not project.is_empty
    assert project.folders == []

    analysis: FolderAnalysis = project.analysis
    assert analysis.lines == {"code": 1, "documentation": 3, "empty": 0, "string": 0}
    assert len(project.modules) == 1
    assert project.modules[0].name == "__init__"
    assert project.data == {
        "num_commits": None,
        "created": None,
        "updated": None,
        "lines": 1,
        "full_lines": 4,
        "urls": None,
        "docstring": "An example Python package for testing purposes",
        "version": "0.0.1",
        "name": "python_example",
    }


def test_javascript_project():
    project = get_js_project()

    assert sorted(project.file_names) == ["index.js", "package.json"]
    assert project.file_paths == [
        os.path.abspath(os.path.join(JS_PROJECT_PATH, "index.js")),
    ]
    assert project.docstring == "An example JavaScript package"
    assert project.version == Version.from_str("1.0.0")
    assert project.name == JS_PROJECT_NAME
    assert project.path == os.path.abspath(JS_PROJECT_PATH)
    assert project.repo is None
    assert not project.is_empty
    assert project.folders == []

    analysis: FolderAnalysis = project.analysis
    assert analysis.lines == {"code": 1, "documentation": 0, "empty": 0, "string": 0}
    assert len(project.modules) == 1
    assert project.modules[0].name == "index"
    assert project.data == {
        "num_commits": None,
        "created": None,
        "updated": None,
        "lines": 1,
        "full_lines": 1,
        "urls": None,
        "docstring": "An example JavaScript package",
        "version": "1.0.0",
        "name": "js_example",
    }


def test_git_project(git_project: Project):
    project = git_project

    assert project.name == "gitrepo"
    assert project.file_names == ["tempfile"]
    assert isinstance(project.repo, Repo)
    analysis: FolderAnalysis = project.analysis
    assert analysis.lines == {"code": 0, "documentation": 0, "empty": 0, "string": 0}
    assert len(project.modules) == 1
    assert project.modules[0].name == "tempfile"
    data = project.data
    assert data["num_commits"] == 1

    assert (data["created"] - datetime.datetime.now(pytz.utc)).total_seconds() < 180
    assert (data["updated"] - datetime.datetime.now(pytz.utc)).total_seconds() < 180


def test_github_project(github_project: Project):
    project = github_project

    assert isinstance(project.repo, Repo)
    assert project.name == "github-project-example"
    analysis: FolderAnalysis = project.analysis
    assert analysis.lines["code"] == 18
    assert analysis.lines["documentation"] == 0
    assert analysis.lines["empty"] == 4
    assert analysis.lines["string"] == 0
    data = project.data
    assert data["urls"] == ["https://github.com/nickderobertis/github-project-example"]



from setuptools import find_packages, setup

import conf

extra_kwargs = {}

entry_points = None
if conf.CONSOLE_SCRIPTS:
    entry_points = dict(console_scripts=conf.CONSOLE_SCRIPTS)

extras_require = None
if conf.OPTIONAL_PACKAGE_INSTALL_REQUIRES:
    extras_require = conf.OPTIONAL_PACKAGE_INSTALL_REQUIRES

long_description = conf.PACKAGE_DESCRIPTION
if conf.PACKAGE_DESCRIPTION.strip().lower() == "auto":
    with open("README.md", "r") as f:
        long_description = f.read()
    extra_kwargs["long_description_content_type"] = "text/markdown"

setup(
    name=conf.PACKAGE_NAME,
    description=conf.PACKAGE_SHORT_DESCRIPTION,
    long_description=long_description,
    author=conf.PACKAGE_AUTHOR,
    author_email=conf.PACKAGE_AUTHOR_EMAIL,
    license=conf.PACKAGE_LICENSE,
    packages=find_packages(),
    include_package_data=True,
    classifiers=conf.PACKAGE_CLASSIFIERS,
    install_requires=conf.PACKAGE_INSTALL_REQUIRES,
    extras_require=extras_require,
    project_urls=conf.PACKAGE_URLS,
    url=conf.PACKAGE_URLS["Code"],
    scripts=conf.SCRIPTS,
    entry_points=entry_points,
    **extra_kwargs
)



from projectreport.finder.base import Finder


class PythonPackageFinder(Finder):
    """
    Finder which automatically looks for projects with Python packages.
    """

    def __init__(self, recursive: bool = True):
        super().__init__(recursive=recursive, required_files=("__init__.py",))



from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from projectreport.analyzer.module import Module
    from projectreport.analyzer.analyzable import Analyzable
    from projectreport.analyzer.folder import Folder

from datetime import datetime

import pygount
from cached_property import cached_property
from git import Commit  # type: ignore


class Analysis:
    _loc: Optional[int] = None
    _full_loc: Optional[int] = None

    def __init__(self, analyzable: "Analyzable"):
        self.git_analysis = GitAnalysis(analyzable)

    @cached_property
    def data(self):
        return dict(
            num_commits=self.git_analysis.num_commits,
            created=self.git_analysis.created,
            updated=self.git_analysis.updated,
            lines=self.loc,
            full_lines=self.full_loc,
            urls=self.git_analysis.urls,
        )

    @property
    def loc(self) -> Optional[int]:
        return self._loc

    @loc.setter
    def loc(self, value: int):
        self._loc = value

    @property
    def full_loc(self) -> Optional[int]:
        return self._full_loc

    @full_loc.setter
    def full_loc(self, value: int):
        self._full_loc = value


class FolderAnalysis(Analysis):
    def __init__(self, folder: "Folder"):
        self.lines = {key: 0 for key in ["code", "documentation", "empty", "string"]}
        super().__init__(folder)

    def __repr__(self):
        return f"<FolderAnalysis(lines={self.lines})>"

    def add_module_analysis(self, analysis: "ModuleAnalysis"):
        for attr in self.lines:
            value = getattr(analysis.source_analysis, attr)
            if value is None:
                value = 0
            self.lines[attr] += value

    def add_subfolder_analysis(self, analysis: "FolderAnalysis"):
        for attr in self.lines:
            value = analysis.lines[attr]
            if value is None:
                value = 0
            self.lines[attr] += value

    @property
    def loc(self) -> int:
        return self.lines["code"]

    @loc.setter
    def loc(self, value: int):
        raise NotImplementedError

    @property
    def full_loc(self) -> int:
        loc = 0
        keys = ["code", "documentation", "empty", "string"]
        for key in keys:
            loc += self.lines[key]
        return loc

    @full_loc.setter
    def full_loc(self, value: int):
        raise NotImplementedError


class ModuleAnalysis(Analysis):
    def __init__(self, module: "Module"):
        self.module = module
        self.source_analysis: pygount.SourceAnalysis = pygount.SourceAnalysis.from_file(
            self.module.path, self.module.package
        )
        self.loc = self.source_analysis.code

        full_loc = 0
        for key in ["code", "documentation", "empty", "string"]:
            full_loc += getattr(self.source_analysis, key, 0)
        self.full_loc = full_loc
        super().__init__(module)


class GitAnalysis:
    def __init__(self, analyzable: "Analyzable"):
        self.ref = analyzable

    @cached_property
    def commits(self) -> Optional[List[Commit]]:
        if not self.has_repo:
            return None
        if self.ref.project is None:  # for mypy
            return None
        commits = [
            commit for commit in self.ref.project.repo.iter_commits(paths=self.ref.path)
        ]
        return commits

    @cached_property
    def num_commits(self) -> Optional[int]:
        if not self.has_repo:
            return None
        return len(self.commits)

    @cached_property
    def created(self) -> Optional[datetime]:
        if not self.commits:
            return None
        return self.commits[-1].committed_datetime

    @cached_property
    def updated(self) -> Optional[datetime]:
        if not self.commits:
            return None
        return self.commits[0].committed_datetime

    @cached_property
    def has_repo(self) -> bool:
        return self.ref.project is not None and self.ref.project.repo is not None

    @cached_property
    def has_remote(self) -> bool:
        if self.ref.project is None:
            return False
        try:
            self.ref.project.repo.remote()
            return True
        except ValueError as e:
            if "Remote named 'origin' didn't exist" in str(e):
                # Repo does not have remote
                return False
            else:
                raise e

    @cached_property
    def urls(self) -> Optional[List[str]]:
        if not self.has_repo or not self.has_remote:
            return None
        if self.ref.project is None:  # for mypy
            return None
        urls = list(self.ref.project.repo.remote().urls)

        # Urls currently have .git on the end, strip
        urls = [url[:-4] if url.endswith(".git") else url for url in urls]
        return urls



import pytest
import semver

from projectreport.version import Version


def test_version_from_semver_str():
    version1 = Version.from_str("3.4.5-pre.2+build.4")
    version2 = Version.from_str("v3.4.5-pre.2+build.4")
    for version in [version1, version2]:
        assert version.major == 3
        assert version.minor == 4
        assert version.patch == 5
        assert version.prerelease == "pre.2"
        assert version.build == "build.4"
        assert version.is_semver
        assert version.semantic_version == semver.VersionInfo.parse(
            "3.4.5-pre.2+build.4"
        )
    assert str(version1) == version1.full_str == "3.4.5-pre.2+build.4"
    assert str(version2) == version2.full_str == "v3.4.5-pre.2+build.4"


def test_version_from_non_semver_str():
    version1 = Version.from_str("3")
    version2 = Version.from_str("v3")
    for version in [version1, version2]:
        assert version.major is None
        assert version.minor is None
        assert version.patch is None
        assert version.prerelease is None
        assert version.build is None
        assert not version.is_semver
        with pytest.raises(ValueError) as excinfo:
            stuff = version.semantic_version
        assert "is_semver=False" in excinfo.exconly()
    assert str(version1) == version1.full_str == "3"
    assert str(version2) == version2.full_str == "v3"


def test_compare_versions():
    version1 = Version.from_str("1.0.0")
    version2 = Version.from_str("1.0.1")
    version3 = Version.from_str("1.0.0")
    not_semver = Version.from_str("blah")
    assert version1 < version2
    assert version1 <= version2
    assert version2 > version1
    assert version2 >= version1
    assert version1 == version3
    assert version1 == semver.VersionInfo.parse("1.0.0")
    assert version1 == "1.0.0"
    assert not version1 == not_semver
    assert not version2 == not_semver
    assert not version3 == not_semver

    for version in [version1, version2, version3]:
        with pytest.raises(ValueError) as excinfo:
            version > not_semver
        assert "is_semver=False, cannot create semver object" in str(excinfo)
        with pytest.raises(ValueError) as excinfo:
            not_semver > version
        assert "cannot compare non-semantic versions" in str(excinfo)

        with pytest.raises(ValueError) as excinfo:
            version >= not_semver
        assert "is_semver=False, cannot create semver object" in str(excinfo)
        with pytest.raises(ValueError) as excinfo:
            not_semver >= version
        assert "cannot compare non-semantic versions" in str(excinfo)

        with pytest.raises(ValueError) as excinfo:
            version < not_semver
        assert "is_semver=False, cannot create semver object" in str(excinfo)
        with pytest.raises(ValueError) as excinfo:
            not_semver < version
        assert "cannot compare non-semantic versions" in str(excinfo)

        with pytest.raises(ValueError) as excinfo:
            version <= not_semver
        assert "is_semver=False, cannot create semver object" in str(excinfo)
        with pytest.raises(ValueError) as excinfo:
            not_semver <= version
        assert "cannot compare non-semantic versions" in str(excinfo)



import logging
import sys

logger = logging.getLogger("project-report")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(levelname)s: %(message)s")

ch = logging.StreamHandler(stream=sys.stdout)
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.addHandler(ch)
